
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  ASYNC.C
*   DESCRIP  :  FENRIS Asynch IO Library
*   DATE     :  July 13, 2000 (my birthday)
*
*
***************************************************************************/

#include "globals.h"

// We use eight (8) processes for the async io manager.  Each disk bin
// is computed as disk % 8 and matched into one of eight bin groups.
// On SMP versions of linux, this should allow us to keep as many
// drive spindles active as possible at one time without eating too
// many server processes to do this, provided the linux kernel
// load balances these processes across processors.

ASYNCH_IO *disk_io_head[8] = { 0, 0, 0, 0, 0, 0, 0 };
ASYNCH_IO *disk_io_tail[8] = { 0, 0, 0, 0, 0, 0, 0 };
ASYNCH_IO *disk_hash[256];  // max is 256 disks

#if (LINUX_SLEEP)
struct semaphore disk_io_sem0 = MUTEX;
struct semaphore disk_io_sem1 = MUTEX;
struct semaphore disk_io_sem2 = MUTEX;
struct semaphore disk_io_sem3 = MUTEX;
struct semaphore disk_io_sem4 = MUTEX;
struct semaphore disk_io_sem5 = MUTEX;
struct semaphore disk_io_sem6 = MUTEX;
struct semaphore disk_io_sem7 = MUTEX;

struct semaphore *io_sem_table[8]={
    &disk_io_sem0, &disk_io_sem1, &disk_io_sem2, &disk_io_sem3,
    &disk_io_sem4, &disk_io_sem5, &disk_io_sem6, &disk_io_sem7
};

#endif

extern void RunAsynchIOQueue(ULONG disk);

void lock_io(ULONG disk)
{
#if (LINUX_SLEEP)
    if (WaitOnSemaphore(io_sem_table[(disk % 8)]) == -EINTR)
       NWFSPrint("lock io was interrupted\n");
#endif
}

void unlock_io(ULONG disk)
{
#if (LINUX_SLEEP)
    SignalSemaphore(io_sem_table[(disk % 8)]);
#endif
}

void process_asynch_io(ULONG disk)
{
   register int i = (disk % 8);
   ASYNCH_IO *list, *io;

   // continue to cycle through this list until we have completely
   // emptied the list.

   while (disk_io_head[i])
   {
      // take the entire list, zero the head and tail and process the list
      // from this context.  this will simulate an alternating A and B list
      // and avoid elevator starvation.

      lock_io(disk);
      list = disk_io_head[i];
      disk_io_head[i] = disk_io_tail[i] = 0;
      unlock_io(disk);

      while (list)
      {
         io = list;
         list = list->next;
         io->next = io->prior = 0;

         if (io->signature != ASIO_SUBMIT_IO)
         {
            NWFSPrint("nwfs:  asynch io process - bad io request cmd-%X s-%X\n",
	              (unsigned)io->command, (unsigned)io->signature);
            io->ccode = ASIO_BAD_SIGNATURE;
            if (io->call_back_routine)
               (io->call_back_routine)(io);
            io->signature = 0;
            continue;
         }

         switch (io->command)
         {
            case ASYNCH_READ:
               io->return_code = pReadDiskSectors(io->disk,
	                                          io->sector_offset,
                                                  io->buffer,
						  io->sector_count,
				                  io->sector_count);
               io->ccode = 0;
               if (!io->return_code)
                  io->ccode = ASIO_IO_ERROR;
               if (io->call_back_routine)
                  (io->call_back_routine)(io);
               io->signature = 0;
               break;

            case ASYNCH_WRITE:
               io->return_code = pWriteDiskSectors(io->disk,
	                                           io->sector_offset,
                                                   io->buffer,
						   io->sector_count,
				                   io->sector_count);
               io->ccode = 0;
               if (!io->return_code)
                  io->ccode = ASIO_IO_ERROR;
               if (io->call_back_routine)
                  (io->call_back_routine)(io);
               io->signature = 0;
               break;

            case ASYNCH_FILL:
               io->return_code = pZeroFillDiskSectors(io->disk,
	                                              io->sector_offset,
						      io->sector_count,
				                      io->sector_count);
               io->ccode = 0;
               if (!io->return_code)
                  io->ccode = ASIO_IO_ERROR;
               if (io->call_back_routine)
                  (io->call_back_routine)(io);
               io->signature = 0;
               break;

            default:
               io->ccode = ASIO_BAD_COMMAND;
               if (io->call_back_routine)
                  (io->call_back_routine)(io);
               io->signature = 0;
               break;
         }
      }
   }
   return;
}

ULONG asynch_io_pending(ULONG disk)
{
   return ((disk_io_head[(disk % 8)]) ? 1 : 0);
}

ASYNCH_IO *remove_io(ULONG disk, ASYNCH_IO *io)
{
    register int i = (disk % 8);

    lock_io(disk);
    if (disk_io_head[i] == io)
    {
       disk_io_head[i] = (void *) io->next;
       if (disk_io_head[i])
	  disk_io_head[i]->prior = NULL;
       else
	  disk_io_tail[i] = NULL;
    }
    else
    {
       io->prior->next = io->next;
       if (io != disk_io_tail[i])
	  io->next->prior = io->prior;
       else
	  disk_io_tail[i] = io->prior;
    }
    io->next = io->prior = 0;
    unlock_io(disk);
    return io;

}

ASYNCH_IO *index_io(ULONG disk, ASYNCH_IO *io)
{
    register int i = (disk % 8);
    ASYNCH_IO *old, *p;

    if (!disk_io_tail[i])
    {
       io->next = io->prior = NULL;
       disk_io_tail[i] = io;
       return io;
    }


    p = disk_io_head[i];
    old = NULL;
    while (p)
    {
       if ((p->disk < io->disk) && (p->sector_offset < io->sector_offset))
       {
	  old = p;
	  p = p->next;
       }
       else
       {
	  if (p->prior)
	  {
	     p->prior->next = io;
	     io->next = p;
	     io->prior = p->prior;
	     p->prior = io;
	     return disk_io_head[i];
	  }
	  io->next = p;
	  io->prior = NULL;
	  p->prior = io;
	  return io;
       }
    }
    old->next = io;
    io->next = NULL;
    io->prior = old;
    disk_io_tail[i] = io;
    return disk_io_head[i];
}

void insert_io(ULONG disk, ASYNCH_IO *io)
{
    lock_io(disk);
    if (io->signature == ASIO_SUBMIT_IO)
    {
       NWFSPrint("nwfs:  asynch io request already active\n");
       unlock_io(disk);
       return;
    }
    disk_io_head[(disk % 8)] = index_io(disk, io);
    io->signature = ASIO_SUBMIT_IO;
    unlock_io(disk);
    return;

}

